@@ -2,7 +2,7 @@
 #
 # The MIT License (MIT)
 #
-# Copyright (c) Tavendo GmbH
+# Copyright (c) Crossbar.io Technologies GmbH
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -26,23 +26,27 @@
 
 from __future__ import absolute_import
 
+import six
 import inspect
 
-import six
+import txaio
+txaio.use_twisted()  # noqa
 
-from twisted.internet.defer import inlineCallbacks
+from twisted.internet.defer import inlineCallbacks, succeed
+
+from autobahn.util import public
+
+from autobahn.websocket.util import parse_url as parse_ws_url
+from autobahn.rawsocket.util import parse_url as parse_rs_url
+
+from autobahn.twisted.websocket import WampWebSocketClientFactory
+from autobahn.twisted.rawsocket import WampRawSocketClientFactory
+
+from autobahn.websocket.compress import PerMessageDeflateOffer, \
+    PerMessageDeflateResponse, PerMessageDeflateResponseAccept
 
 from autobahn.wamp import protocol
 from autobahn.wamp.types import ComponentConfig
-from autobahn.websocket.util import parse_url
-from autobahn.twisted.websocket import WampWebSocketClientFactory
-
-# new API
-# from autobahn.twisted.connection import Connection
-
-import txaio
-txaio.use_twisted()
-
 
 __all__ = [
     'ApplicationSession',
@@ -52,7 +56,8 @@ __all__ = [
     'Service',
 
     # new API
-    'Session'
+    'Session',
+    # 'run',  # should probably move this method to here? instead of component
 ]
 
 try:
@@ -63,11 +68,19 @@ except (ImportError, SyntaxError):
     __all__.pop(__all__.index('Service'))
 
 
+@public
 class ApplicationSession(protocol.ApplicationSession):
     """
     WAMP application session for Twisted-based applications.
+
+    Implements:
+
+        * :class:`autobahn.wamp.interfaces.ITransportHandler`
+        * :class:`autobahn.wamp.interfaces.ISession`
     """
 
+    log = txaio.make_logger()
+
 
 class ApplicationSessionFactory(protocol.ApplicationSessionFactory):
     """
@@ -79,7 +92,10 @@ class ApplicationSessionFactory(protocol.ApplicationSessionFactory):
     The application session class this application session factory will use. Defaults to :class:`autobahn.twisted.wamp.ApplicationSession`.
     """
 
+    log = txaio.make_logger()
 
+
+@public
 class ApplicationRunner(object):
     """
     This class is a convenience tool mainly for development and quick hosting
@@ -91,14 +107,21 @@ class ApplicationRunner(object):
 
     log = txaio.make_logger()
 
-    def __init__(self, url, realm, extra=None, serializers=None, ssl=None, proxy=None):
+    def __init__(self,
+                 url,
+                 realm=None,
+                 extra=None,
+                 serializers=None,
+                 ssl=None,
+                 proxy=None,
+                 headers=None):
         """
 
         :param url: The WebSocket URL of the WAMP router to connect to (e.g. `ws://somehost.com:8090/somepath`)
-        :type url: unicode
+        :type url: str
 
         :param realm: The WAMP realm to join the application session to.
-        :type realm: unicode
+        :type realm: str
 
         :param extra: Optional extra configuration to forward to the application component.
         :type extra: dict
@@ -118,10 +141,14 @@ class ApplicationRunner(object):
 
         :param proxy: Explicit proxy server to use; a dict with ``host`` and ``port`` keys
         :type proxy: dict or None
+
+        :param headers: Additional headers to send (only applies to WAMP-over-WebSocket).
+        :type headers: dict
         """
         assert(type(url) == six.text_type)
         assert(realm is None or type(realm) == six.text_type)
         assert(extra is None or type(extra) == dict)
+        assert(headers is None or type(headers) == dict)
         assert(proxy is None or type(proxy) == dict)
         self.url = url
         self.realm = realm
@@ -129,16 +156,34 @@ class ApplicationRunner(object):
         self.serializers = serializers
         self.ssl = ssl
         self.proxy = proxy
+        self.headers = headers
 
-    def run(self, make, start_reactor=True):
+        # this if for auto-reconnection when Twisted ClientService is avail
+        self._client_service = None
+        # total number of successful connections
+        self._connect_successes = 0
+
+    @public
+    def stop(self):
+        """
+        Stop reconnecting, if auto-reconnecting was enabled.
+        """
+        self.log.debug('{klass}.stop()', klass=self.__class__.__name__)
+        if self._client_service:
+            return self._client_service.stopService()
+        else:
+            return succeed(None)
+
+    @public
+    def run(self, make, start_reactor=True, auto_reconnect=False, log_level='info'):
         """
         Run the application component.
 
-        :param make: A factory that produces instances of :class:`autobahn.asyncio.wamp.ApplicationSession`
+        :param make: A factory that produces instances of :class:`autobahn.twisted.wamp.ApplicationSession`
            when called with an instance of :class:`autobahn.wamp.types.ComponentConfig`.
         :type make: callable
 
-        :param start_reactor: if True (the default) this method starts
+        :param start_reactor: When ``True`` (the default) this method starts
            the Twisted reactor and doesn't return until the reactor
            stops. If there are any problems starting the reactor or
            connect()-ing, we stop the reactor and raise the exception
@@ -157,32 +202,68 @@ class ApplicationRunner(object):
             from twisted.internet import reactor
             txaio.use_twisted()
             txaio.config.loop = reactor
-            txaio.start_logging(level='info')
+            txaio.start_logging(level=log_level)
 
-        isSecure, host, port, resource, path, params = parse_url(self.url)
-
-        # factory for use ApplicationSession
-        def create():
-            cfg = ComponentConfig(self.realm, self.extra)
-            try:
-                session = make(cfg)
-            except Exception as e:
-                if start_reactor:
-                    # the app component could not be created .. fatal
-                    self.log.error("{err}", err=e)
-                    reactor.stop()
-                else:
-                    # if we didn't start the reactor, it's up to the
-                    # caller to deal with errors
+        if callable(make):
+            # factory for use ApplicationSession
+            def create():
+                cfg = ComponentConfig(self.realm, self.extra)
+                try:
+                    session = make(cfg)
+                except Exception:
+                    self.log.failure('ApplicationSession could not be instantiated: {log_failure.value}')
+                    if start_reactor and reactor.running:
+                        reactor.stop()
                     raise
-            else:
-                return session
+                else:
+                    return session
+        else:
+            create = make
 
-        # create a WAMP-over-WebSocket transport client factory
-        transport_factory = WampWebSocketClientFactory(create, url=self.url, serializers=self.serializers, proxy=self.proxy)
+        if self.url.startswith(u'rs'):
+            # try to parse RawSocket URL ..
+            isSecure, host, port = parse_rs_url(self.url)
 
-        # supress pointless log noise like
-        # "Starting factory <autobahn.twisted.websocket.WampWebSocketClientFactory object at 0x2b737b480e10>""
+            # use the first configured serializer if any (which means, auto-choose "best")
+            serializer = self.serializers[0] if self.serializers else None
+
+            # create a WAMP-over-RawSocket transport client factory
+            transport_factory = WampRawSocketClientFactory(create, serializer=serializer)
+
+        else:
+            # try to parse WebSocket URL ..
+            isSecure, host, port, resource, path, params = parse_ws_url(self.url)
+
+            # create a WAMP-over-WebSocket transport client factory
+            transport_factory = WampWebSocketClientFactory(create, url=self.url, serializers=self.serializers, proxy=self.proxy, headers=self.headers)
+
+            # client WebSocket settings - similar to:
+            # - http://crossbar.io/docs/WebSocket-Compression/#production-settings
+            # - http://crossbar.io/docs/WebSocket-Options/#production-settings
+
+            # The permessage-deflate extensions offered to the server ..
+            offers = [PerMessageDeflateOffer()]
+
+            # Function to accept permessage_delate responses from the server ..
+            def accept(response):
+                if isinstance(response, PerMessageDeflateResponse):
+                    return PerMessageDeflateResponseAccept(response)
+
+            # set WebSocket options for all client connections
+            transport_factory.setProtocolOptions(maxFramePayloadSize=1048576,
+                                                 maxMessagePayloadSize=1048576,
+                                                 autoFragmentSize=65536,
+                                                 failByDrop=False,
+                                                 openHandshakeTimeout=2.5,
+                                                 closeHandshakeTimeout=1.,
+                                                 tcpNoDelay=True,
+                                                 autoPingInterval=10.,
+                                                 autoPingTimeout=5.,
+                                                 autoPingSize=4,
+                                                 perMessageCompressionOffers=offers,
+                                                 perMessageCompressionAccept=accept)
+
+        # supress pointless log noise
         transport_factory.noisy = False
 
         # if user passed ssl= but isn't using isSecure, we'll never
@@ -212,8 +293,6 @@ class ApplicationRunner(object):
             from twisted.internet.endpoints import TCP4ClientEndpoint
             client = TCP4ClientEndpoint(reactor, host, port)
 
-        d = client.connect(transport_factory)
-
         # as the reactor shuts down, we wish to wait until we've sent
         # out our "Goodbye" message; leave() returns a Deferred that
         # fires when the transport gets to STATE_CLOSED
@@ -227,9 +306,49 @@ class ApplicationRunner(object):
         # when our proto was created and connected, make sure it's cleaned
         # up properly later on when the reactor shuts down for whatever reason
         def init_proto(proto):
+            self._connect_successes += 1
             reactor.addSystemEventTrigger('before', 'shutdown', cleanup, proto)
             return proto
 
+        use_service = False
+        if auto_reconnect:
+            try:
+                # since Twisted 16.1.0
+                from twisted.application.internet import ClientService
+                from twisted.application.internet import backoffPolicy
+                use_service = True
+            except ImportError:
+                use_service = False
+
+        if use_service:
+            # this code path is automatically reconnecting ..
+            self.log.debug('using t.a.i.ClientService')
+
+            default_retry = backoffPolicy()
+
+            if False:
+                # retry policy that will only try to reconnect if we connected
+                # successfully at least once before (so it fails on host unreachable etc ..)
+                def retry(failed_attempts):
+                    if self._connect_successes > 0:
+                        return default_retry(failed_attempts)
+                    else:
+                        self.stop()
+                        return 100000000000000
+            else:
+                retry = default_retry
+
+            self._client_service = ClientService(client, transport_factory, retryPolicy=retry)
+            self._client_service.startService()
+
+            d = self._client_service.whenConnected()
+
+        else:
+            # this code path is only connecting once!
+            self.log.debug('using t.i.e.connect()')
+
+            d = client.connect(transport_factory)
+
         # if we connect successfully, the arg is a WampWebSocketClientProtocol
         d.addCallback(init_proto)
 
@@ -574,7 +693,7 @@ if service:
             """
             Setup the application component.
             """
-            is_secure, host, port, resource, path, params = parse_url(self.url)
+            is_secure, host, port, resource, path, params = parse_ws_url(self.url)
 
             # factory for use ApplicationSession
             def create():
@@ -605,17 +724,28 @@ if service:
 
 # new API
 class Session(ApplicationSession):
+    # shim that lets us present pep8 API for user-classes to override,
+    # but also backwards-compatible for existing code using
+    # ApplicationSession "directly".
+
+    # XXX note to self: if we release this as "the" API, then we can
+    # change all internal Autobahn calls to .on_join() etc, and make
+    # ApplicationSession a subclass of Session -- and it can then be
+    # separate deprecated and removed, ultimately, if desired.
 
     def onJoin(self, details):
         return self.on_join(details)
 
+    # XXX def onChallenge(self, )
+    # XXX def onOpen ??
+
     def onLeave(self, details):
         return self.on_leave(details)
 
     def onDisconnect(self):
         return self.on_disconnect()
 
-    def on_join(self):
+    def on_join(self, details):
         pass
 
     def on_leave(self, details):
